import gc
import os
import traceback as tb
from collections import defaultdict
from datetime import datetime as dt
from multiprocessing import Pool

import pandas as pd
import psutil
from polly.auth import Polly
from polly.omixatlas import OmixAtlas
from sqlalchemy import create_engine
from tqdm import tqdm

from ipc.utility.helper import chunks
from ipc.validator import sample_metadata_validator
from ipc.validator import dataset_metadata_validator
from ipc.validator.settings import DATASET_ERRORS_TABLE, DB_NAME, REPOS_FOR_DATASET_LEVEL_VALIDATION, \
    REPOS_FOR_SAMPLE_LEVEL_VALIDATION, RUN_DATASET_LEVEL_VALIDATIONS, RUN_SAMPLE_LEVEL_VALIDATIONS, SAMPLE_ERRORS_TABLE

# Obtain authentication tokens
try:
    if 'AUTH_TOKEN' in os.environ:
        AUTH_TOKEN = os.environ['AUTH_TOKEN']
        # print('Yes its there')
    else:
        raise Exception
except Exception as e:
    print(f'Polly-python authentication token missing.')
    print(tb.format_exc())
    exception_type, exception_object, exception_traceback = sys.exc_info()
    filename = exception_traceback.tb_frame.f_code.co_filename
    line_number = exception_traceback.tb_lineno
    print("Exception type: ", exception_type)
    print("File name: ", filename)
    print("Line number: ", line_number)

# AUTH_TOKEN = 'eyJjdHkiOiJKV1QiLCJlbmMiOiJBMjU2R0NNIiwiYWxnIjoiUlNBLU9BRVAifQ.eAxEF4Qy8PZPMx4hVBPSl4Amd-OOQA0Iy57YeyQAGthLcy1_vW5THhUJ1bhlNRDwBeg_8svM30evW261113ugFFodYkYZAmNyVnzH7drs35JZ4iM0Daa9_afxMxJhgJtz7Y0uWCakMhFkZtKBsvX96fa57MnrlkJ_h4yzLTNC7sW_Wk4a2O9KMKbgC_EcwhBqZSaoCRt3CsTfudSoFgBe3J5e1_Ouq6PUbx7cGta2ozrc3kW8-bj6XbMnHApG09Sn090ilflNs8mtGONfz22turuoC_Q9AypAs_ZRgqBkoPHFnHn-BdZ3Y6qJK0AtIqlAHBpVs8GFDOpkeEMFH832w.jnil8O3I6DgExVTs.JLhRMl2_g4zKt5XvSs0r2xoTJDXndJ4z58IyR7CTOwe2H_OktgMDXd1SNvz1scLfP6t8BNfwNv5xnUjbRAdBk3y0bSXlcttA5vxnw3dA8yQhxgDjMy7ZNhqxDO9XwuNOr8XYQV1ob7clipciWplI4RJx3BdJOcWEuaKyQRN9VvG4qdQ6tHDeEInjsc5CX9tWK7oLsIb0rrELU2x3NEXb0Ch3mQHSbbUG9Azvu7I6zpB7uI7AW-nnfR1DWf2V4_gRo3SHn2pgGsEvOkx-AvDKPyjGAcGClthO0Zp-DOI-Yf_j0ukDzSS3GfQvJWOiW-0EsJeGHC-SnEdM4lQIfW8iQZd2upshSu_zo6Jap6Fl2KLZOjPsTH7KuPx99JMPeZBH_ZJSP7-3g6H8nX3H2qhMuZBq3qtGiz2TZUW2TaVOG3QCLBg3yKXPmbfqVdKIFLK110a3gUWNyQCx89o4AKzC3V-D1n0ZlnesiN3qOgeR602dmwLjsVNTCPy0x8P4oJDMc7vZ9Usna2k33yzT8948SbFRShbY_ZnamYM33zfKqdWS6BF6Fmi6gcRm3IA9abAa8YdbwjOONmDUORd2g1ZyzuzWTZdPYni-D4_dVgPNLGnWWlQsqFuvL_G3UitrQ_5Qs0Hm0x1SpVI5rSP8_PsM_GlOojlQW2hFss45yvvCglT7D5pA94zhnm2r2AtbNQ9n9PxpLpmZJb2G0oQZLsPTD9IPLAASaN-V_Q0DOL4pHKkPd1Bzk_IZsoIktW7ryGOqZ5twHDbspLgaa6OMEX2PS1L4sT64-_NGZl0o6igmMnHODzZSEBKXzHk1QESfHtjsyrRoX2iNYW2KsuYXcRpWhdD0W4NFLXYz7ZBomIChFY8lQOrGlbkgg_6D81Tc76mwo7Dat_cgCNTV_kQUsA7PuNgw8ukC6qEJ6g7rrvlrHoACYfTGZvgOIKpr6JuwHL2LyB8dZLnI2c6EftZJ6NymXSDJi8dUV7ydY4LzGmNM-cSi1jz28GvVzuG-7Qs-y4MfC6oCuFeWk76_3y3Rl3h7FhYMd03OXLP_sb3JXU1AuSzRGvDBTNlDytG419XXUHwgjK4E-FC-5uhAwUWTYcjLjr-Qg2qctGGctAiEvOnmdoKoGITm1OkbPWk5CnKZyzQq5uwSuJwciYRL1w7QUiu5stXKQ1Gzsf6alr2srHvr1pgbqlKupgm9iHoBqvkdAobj2lr78W6lBKpR2mHRm6sSlHFvBe-N_62b4Jvr5hkjiqzcQtYIqTrRPqc5dQdXp0VAT8jEX57422y9mwS-p758lsJd4zKrkRxYffNKZ-Q7sH6YXCnKQVnpitQg29E.iBod3im8hR4D2BdMxaCWBw'

Polly.auth(AUTH_TOKEN)
# Defining OMIXATLAS object
OMIXATLAS = OmixAtlas()


def get_query_results(query):
    """
    Given a query for Polly-python, return the resultant DF
    Args:
        query: SQL Query

    Returns:
        DataFrame: Response to the query
    """
    return OMIXATLAS.query_metadata(query, query_api_version="v2")


"""----Functions for repo-wise dataset level metadata validation----"""


def dataset_metadata_validator_construct_args_for_pool(repos):
    """
    Function to create arguments for pool of workers to run
    Args:
        repos:

    Returns:

    """
    args_for_pool = defaultdict(list)
    for repo in repos:
        print(f'Getting datasets for: {repo}')
        t_start = dt.now()
        dataset_metadatas = get_query_results(f'SELECT * FROM {repo}.datasets')
        print(f'Actual time taken to run the query: {dt.now() - t_start}')

        print(f'Getting schema for: {repo}')
        t_start = dt.now()
        desc_df = get_query_results(f'DESCRIBE {repo}.datasets')
        print(f'Actual time taken to run the query: {dt.now() - t_start}')

        for val in dataset_metadatas.to_dict('records'):
            args_for_pool[repo].append((repo, desc_df, val))
    return args_for_pool


def check_datasets_in_repo_for_errors(repos):
    """

    Args:
        repo:

    Returns:

    """
    compiled_error_df = pd.DataFrame(columns=['Field', 'Error Message', 'Error Type', 'Repo',
                                              'dataset_id', 'key'])
    try:
        num_repos = len(repos)
        args_for_pool = dataset_metadata_validator_construct_args_for_pool(repos)
        print(f'RAM Usage: {psutil.virtual_memory().percent}%')
        ctr = 1
        print(f'Compiling errors...')
        print(f'\nWill use {psutil.cpu_count()} cores at my disposal...\n')
        chunk_size = 10000
        for repo in repos:
            chunk_no = 1
            print(f'Working on repo: {repo} | Count: {ctr}/{num_repos}')
            total_chunks = len(args_for_pool[repo]) // chunk_size
            tstart = dt.now()
            for args in chunks(args_for_pool[repo], chunk_size):
                print(f'Chunk: {chunk_no}/{total_chunks}')
                process_pool = Pool(processes=psutil.cpu_count(logical=False))
                error_df_results = process_pool.starmap(dataset_metadata_validator.check_per_item,
                                                        tqdm(args, total=len(args)))

                # error_df_results = pd.concat(error_df_results, ignore_index=True)
                error_dfs = [res[0] for res in error_df_results]
                compiled_error_df = pd.concat([compiled_error_df, pd.concat(error_dfs, ignore_index=True)],
                                              ignore_index=True)
                del error_df_results, error_dfs
                gc.collect()
                chunk_no += 1
                print(f'RAM Usage: {psutil.virtual_memory().percent}%')
            print(f'Time Taken: {dt.now() - tstart}\n')
            ctr += 1
        compiled_error_df.reset_index(inplace=True, drop=True)
        print(f'RAM Usage: {psutil.virtual_memory().percent}%')
        return compiled_error_df
    except Exception as err:
        print(f'Error when collecting errors: {tb.format_exc()}')
        exception_type, exception_object, exception_traceback = sys.exc_info()
        filename = exception_traceback.tb_frame.f_code.co_filename
        line_number = exception_traceback.tb_lineno
        print("Exception type: ", exception_type)
        print("File name: ", filename)
        print("Line number: ", line_number)
    return pd.DataFrame(columns=['Field', 'Error Message', 'Error Type', 'Repo',
                                 'dataset_id', 'key'])


"""----Functions for repo-wise sample level metadata validation----"""


def sample_metadata_validator_construct_args_for_pool(repos):
    """

    Args:
        repos:

    Returns:

    """
    args_for_pool = defaultdict(list)
    for repo in repos:
        print(f'Getting samples for: {repo}')
        t_start = dt.now()
        if repo == 'sc_data_lake':
            samples = get_query_results(f'SELECT * FROM {repo}.samples_singlecell')
        else:
            samples = get_query_results(f'SELECT * FROM {repo}.samples')
        print(f'Actual time taken to run the query: {dt.now() - t_start}')
        if repo == 'sc_data_lake':
            desc_df = get_query_results(f'DESCRIBE {repo}.samples_singlecell')
        else:
            desc_df = get_query_results(f'DESCRIBE {repo}.samples')
        for val in samples.to_dict('records'):
            args_for_pool[repo].append((repo, desc_df, val))
    return args_for_pool


def check_samples_in_repo_for_errors(repos):
    """
    Given a list of repos, go repo by repo and collect errors
    Args:
        repos:list of repos

    Returns:
        pandas DataFrame: containing the errors collected from all the repos
    """
    compiled_error_df = pd.DataFrame(columns=['Field', 'Error Message', 'Error Type', 'Repo',
                                              'src_dataset_id', 'data_id', 'sample_id'])
    try:
        num_repos = len(repos)
        args_for_pool = sample_metadata_validator_construct_args_for_pool(repos)
        print(f'RAM Usage: {psutil.virtual_memory().percent}%')
        ctr = 1
        print(f'Compiling errors...')
        print(f'\nWill use {psutil.cpu_count()} cores at my disposal...\n')
        chunk_size = 10000
        for repo in repos:
            chunk_no = 1
            print(f'Working on repo: {repo} | Count: {ctr}/{num_repos}')
            total_chunks = len(args_for_pool[repo]) // chunk_size
            tstart = dt.now()
            for args in chunks(args_for_pool[repo], chunk_size):
                print(f'Chunk: {chunk_no}/{total_chunks}')
                process_pool = Pool(processes=psutil.cpu_count(logical=False))
                error_df_results = process_pool.starmap(sample_metadata_validator.check_per_item,
                                                        tqdm(args, total=len(args)))

                # error_df_results = pd.concat(error_df_results, ignore_index=True)
                error_dfs = [res[0] for res in error_df_results]
                compiled_error_df = pd.concat([compiled_error_df, pd.concat(error_dfs, ignore_index=True)],
                                              ignore_index=True)
                del error_df_results, error_dfs
                gc.collect()
                chunk_no += 1
                print(f'RAM Usage: {psutil.virtual_memory().percent}%')
            print(f'Time Taken: {dt.now() - tstart}\n')
            ctr += 1
        compiled_error_df.reset_index(inplace=True, drop=True)
        print(f'RAM Usage: {psutil.virtual_memory().percent}%')
        return compiled_error_df
    except Exception as err:
        print(f'Error when collecting errors: {tb.format_exc()}')
        exception_type, exception_object, exception_traceback = sys.exc_info()
        filename = exception_traceback.tb_frame.f_code.co_filename
        line_number = exception_traceback.tb_lineno
        print("Exception type: ", exception_type)
        print("File name: ", filename)
        print("Line number: ", line_number)
    return pd.DataFrame(columns=['Field', 'Error Message', 'Error Type', 'Repo',
                                 'src_dataset_id', 'data_id', 'sample_id'])


if __name__ == '__main__':
    # Run the error collection routine
    try:
        # Initialize DB connection and connect
        engine = create_engine(f'sqlite:///ipc/validator/{DB_NAME}')
        # Create an SQLAlchemy connection tob an SQLite object with name: DB_NAME
        conn = engine.connect()
        if RUN_DATASET_LEVEL_VALIDATIONS:
            """Running Dataset Level Error Collection"""
            print(f'Collecting errors in dataset metadata from repos: {REPOS_FOR_DATASET_LEVEL_VALIDATION}')
            # Get the compiled error table.
            compiled_dataset_level_errors_df = check_datasets_in_repo_for_errors(REPOS_FOR_DATASET_LEVEL_VALIDATION)
            if not compiled_dataset_level_errors_df.empty:
                # If the DB exists, append to the table.
                compiled_dataset_level_errors_df = compiled_dataset_level_errors_df.astype(str)
                compiled_dataset_level_errors_df.to_sql(DATASET_ERRORS_TABLE, conn, if_exists='replace', chunksize=1000)
                print(f'\nErrors added to table: {DATASET_ERRORS_TABLE}\nDB: {DB_NAME}')
            else:
                print(f'Error compiling errors')

        if RUN_SAMPLE_LEVEL_VALIDATIONS:
            """Running Sample Level Error Collection"""
            compiled_sample_level_errors_df = check_samples_in_repo_for_errors(REPOS_FOR_SAMPLE_LEVEL_VALIDATION)
            if not compiled_sample_level_errors_df.empty:
                # If the table exists, append to the table.
                compiled_sample_level_errors_df = compiled_sample_level_errors_df.astype(str)
                compiled_sample_level_errors_df.to_sql(SAMPLE_ERRORS_TABLE, conn, if_exists='replace', chunksize=1000)
                print(f'\nErrors added to table: {SAMPLE_ERRORS_TABLE}\nDB: {DB_NAME}')
            else:
                print(f'Error compiling errors')

        # Close the DB connection
        conn.close()
    except Exception as e:
        print(tb.format_exc())
        exception_type, exception_object, exception_traceback = sys.exc_info()
        filename = exception_traceback.tb_frame.f_code.co_filename
        line_number = exception_traceback.tb_lineno
        print("Exception type: ", exception_type)
        print("File name: ", filename)
        print("Line number: ", line_number)
